-- Only load once
if WRL and WRL.loading then goto EOF end

-- Avoid coroutine table overrides
local coroutine = coroutine
local http_request = http.request
local openURL = openURL

WRL.Crash = {}

local function stripDocPrefix(str)
    return str:gsub(".-/Documents/", "")
end

WRL.CrashReport = class()

function WRL.CrashReport:init(errMsg)
    self.report = {
        project = WRL.ProjectName,
        date = os.date("!%Y-%m-%d %T") .. " (UTC)",
        error = (errMsg and stripDocPrefix(errMsg)) or "nil"
    }
    
    local version = "Local"
    do
        -- Get the version of the running project
        local verAsset = asset["webrepo_version"]
        if verAsset then
            local f = io.open(verAsset.path, "r")
            version = f:read("*a")
            f:close()
        end
        self.report.version = version
        
        --[[
        if verAsset then
            -- Compare project checksum to detect modifications
            addToReport("Modified: <unimplemented>")
        end
        ]]
    end
end

function WRL.CrashReport:dumpCoroutine(co, name, crashed, offset)
    self.report.coroutines = self.report.coroutines or {}
    
    -- Add our new coroutine dump
    local dump = {}
    table.insert(self.report.coroutines, dump)
    
    dump.name = name
    dump.status = coroutine.status(co)
    dump.crashed = crashed or nil
    dump.callstack = {}

    local i = offset or 0
    while true do
        local info = debug.getinfo(co, i)
        if info == nil then break end
        
        -- Create a new stack frame
        local frame = {}
        table.insert(dump.callstack, frame)
            
        -- Add thread file:line
        frame.location = {
            file = stripDocPrefix(info.short_src),
            line = info.currentline,
            func = info.name or "?"
        }
            
        -- Add locals to report
        if debug.getlocal(co, i, 1) then
            frame.locals = {}
            local j = 1
            local success
            while true do
                local k, v = debug.getlocal(co, i, j)
                if not k then break end
                success, v = pcall(tostring, v)
                if not success then
                    v = "<" .. type(v) .. ">"
                end
                if string.len(v) > 64 then
                    v = v:sub(1, 64) .. "..."
                end
                if v:match("[^%g%s]") then
                    v = "<binary string>"
                end
                frame.locals[k] = v
                j = j + 1
            end
        end
            
        -- Add upvalues to report
        if info.nups > 0 then
            frame.upvalues = {}
            local j = 1
            local success
            while true do
                local k, v = debug.getupvalue(info.func, j)
                if not k then break end
                success, v = pcall(tostring, v)
                if not success then
                    v = "<" .. type(v) .. ">"
                end
                if string.len(v) > 64 then
                    v = v:sub(1, 64) .. "..."
                end
                if v:match("[^%g%s]") then
                    v = "<binary string>"
                end
                frame.upvalues[k] = v
                j = j + 1
            end
        end
            
        i = i + 1
    end
end

function WRL.CrashReport:send()
    local jsonState = {
        indent = true,
        level = 1,
        keyorder = {
            "project",
            "date",
            "version",
            "error",
            "coroutines",
            "trace"
        }
    }

    -- Stop drawing
    _G.draw = nil
    
    local function initParameters(uploaded)
        parameter.clear()
        if not uploaded then
            parameter.text("Notes")
            parameter.text("Reporter", "anonymous")
            parameter.action("Send crash report", function()
                -- Create the payload
                if #Notes == 0 then Notes = "N/A" end
                Notes = Notes:gsub("…", "...") -- Replace weird ellipses char...
                
                self.report.notes = Notes
                self.report.reporter = Reporter
                
                print("Sending. Please wait...")
                
                -- Upload log trace
                WRL.API.upload("crashtrace", nil, WRL.Trace.Get(), function(traceurl)
        
                    -- Upload crash report
                    WRL.API.upload("crashreport", nil, json.encode(self.report, jsonState), function(reporturl)
                        
                        -- Register the crash report
                        http_request("https://codeawebrepo.co.uk/api/crash", function(_, status)
                            if status == 200 then
                                print("Report sent! Thank you for reporting your issue.")
                                initParameters(true)
                            else
                                print("Failed to send. Please try again.")
                            end
                        end, function()
                            print("Failed to send. Please try again.")
                        end, {
                            method = "POST",
                            headers = {
                                ["Content-Type"] = "application/json"
                            },
                            data = json.encode({
                                report = reporturl:match("uploads/(.*)"),
                                trace = traceurl:match("uploads/(.*)"),
                                project = self.report.project,
                                version = self.report.version,
                                date = self.report.date,
                                reporter = Reporter,
                                notes = Notes
                            },{indent=true})
                        })
                    end)    
                end)
            end)
        end
        parameter.action("Copy report to clipboard", function()
            pasteboard.copy("--[==[\n" .. json.encode(self.report, jsonState) .. "\n]==]")
            print("Copied to clipboard")
        end)
        if uploaded then
            parameter.action("Open report in browser", function()
                local url = require("socket").url
                local date = url.escape(self.report.date)
                openURL("https://codeawebrepo.co.uk/crashreports/" .. self.report.project .. "/" .. self.report.version .. "/" .. date)
            end)
        end
    end
        
    viewer.mode = OVERLAY
    initParameters()
        
    --output.clear()
    print("If you would like to upload your crash report please press the button above.\n\nPlease provide any useful info you may have (what you were doing at the time, any modifications etc.) and your forum name if you wish.")
    
    
    local onScreenMessage = "Crashed!\n\t" .. self.report.error
    _G.tween.update = function()
        background(132, 30, 30)
            
        -- Setup text
        textWrapWidth(WIDTH-20)
        textMode(CENTER)
        textAlign(LEFT)
        fontSize(24)
        fill(255)
        font("Futura-MediumItalic")
            
        -- Draw report
        text(onScreenMessage, WIDTH/2, HEIGHT/2)
    end
end

function WRL.CrashHandler(msg)
    local report = WRL.CrashReport(msg)
    report:dumpCoroutine(coroutine.running(), nil, true, 3)
    report:send()
    return msg
end

function WRL.CrashableCall(fn, ...)
    return xpcall(fn, WRL.CrashHandler, ...)
end

::EOF::