-- Only load once
if WRL and WRL.loading then goto EOF end

-- Options:
--  version     - Specify the version of the dependency to load.
--  loadmain    - Loads the main file of the project/dependency
--  import      - Makes the env of the dependency available in the dependent project.
--  export      - Copies the env of the dependency to the dependent project's env.
--  deps        - Dependencies to be loaded prior to the current project.
--  devdeps     - Dependencies to be loaded when the project is launched from Codea
--  loaddev     - Forces devdeps to be used as dependencies.
--                  (default for the launching project)

--[[
@module: WebRepoLib
@desc: WebRepoLib makes accessing and using community libraries a breeze.
---<h3>Getting Started</h3>Everything revolves around the `WRL.lua` file/tab. When a project contains this file and adds WebRepoLib as a dependency, the magic starts.

`WRL.lua` describes your project's dependencies and how you intend to use them.
To add a dependency, define the `deps` table in `WRL.lua` and add your dependency to it:
```
deps = {
    "Codea+thread" -- Import from WebRepo.
}
```
When adding a new file to a project using WebRepoLib, when launching you will see a dependency error asking you to close and re-open your project. Please do as it asks, return to Codea's project browser and re-open the project.  WebRepoLib detects new files and adds guards to ensure they are loaded at the correct time. Only by re-opening your project will those changes be visible. <b>DO NOT</b> remove the load guards from projects where WebRepoLib is used.
These load guards look like this:
```
----------------- DO NOT REMOVE OR MOVE -----------------
-- Ensure Codea doesn't load this file automatically
-- This MUST be at the top of this file!
if WRL and not WRL.loading then return end
--------------- END DO NOT REMOVE OR MOVE ---------------
```
---<h3>Configuration</h3>Each included dependency can be configured in different ways.
This is done by adding a configuration table to a dependency:
```
deps = {
    { "Codea+thread", { ... } }
}
```

<b>Versions:</b>
By default, WebRepoLib will always use the latest version of any dependency available on WebRepo.
You can also fix the version you with to use by including a version in the config table:
```
deps = {
    { "Codea+thread", { version = "1.1.0" } }
}
```
<b>Local projects:</b>
By prepending your dependency's name with a '~' you can load dependencies from your own Codea folder on your device.
This is most useful during development with multiple dependent projects.
```
deps = {
    "~Codea+thread" -- Import local 'Codea+thread' project
}
```
<b>Access:</b>
By default, a dependency is only available to the project that has included it. This is done by making the dependency's environment accessible via the current project's environment's metatable. This allows projects to maintain their own environments without directly manipulating the environment of others.

This behaviour can be changed using the `export` flag:
```
deps = {
    { "Codea+thread" { export = true } }
}
```When `export == true` the dependency's environment is copied into the including project. This allows dependencies to be made available further up the chain.

This is useful when a dependency defines callback functions used by Codea's runtime, e.g. setup(), draw() etc.

<b>loadmain:</b>
By default, only the project being launched by Codea will have its `Main.lua` file loaded.
By setting this flag in a dependency config, you are telling WebRepoLib that it should also load the dependency's `Main.lua` file as well.

WebRepo itself uses the same technique to ensure it runs correctly:
```
deps = {
    -- Always run the latest available WebRepo release.
    {"WebRepoImpl", {
        export = true, -- Export WebRepoImpl globals into our env.
        loadmain = true -- Load WebRepoImpl's 'Main' tab.
    }}
}
```
---<h3>Development dependencies</h3>Sometimes, you may need to include additional dependencies during development but not when your project is used as a dependency. To do this we use the `devdeps` table. Dependencies added to this table are only loaded when you are launching the project directly, and not as a dependency to another project.

The `devdeps` table is useful for things like docs and debuggers etc. and behaves exactly the same way as the `deps` table. This is how we include the codea+docs dependency used to display these docs:
```
devdeps = {
    { "Codea+docs" }
}
```
]]

-- For url escaping
local socket = require("socket")

-- Make a few functions locals
--local TaskBatch = WRL.TaskBatch
local addTask = function(...) return WRL.TaskManager:add(...) end
local gate = function(...) return WRL.TaskManager:gate(...) end
--local taskLoopYield = WRL.Tasks.taskLoopYield
--local delayUntil = WRL.Tasks.delayUntil
local createInternalEnv = WRL.Env.createInternal
local createExternalEnv = WRL.Env.createExternal
local importEnv = WRL.Env.import

-- print() with a prefix.
local _Gprint = print
local function print(...)
    _Gprint("[WRL]", ...)
end

----------------------------------------
-- Tasks
----------------------------------------
local TASK_kickoff
local TASK_validateProject
local TASK_installProject
local TASK_loadDependencies
local TASK_loadProject
local TASK_loadDeferredDependency
local TASK_go

local ENVS = {}
local manifest

TASK_kickoff = function()
    
    -- We're now doing our manual loading
    WRL.loading = true
    
    -- Current project
    local meta = {
        name = (asset.name):sub(1, -7),
        version = "__LOCAL__",
        loadmain = true, -- allows Main.lua to load,
        loaddev = true
    }
    
    -- Last task to be run should be us importing the
    -- current project's env & executing the setup() function.
    addTask(TASK_go, meta)
    
    -- Load the current project
    addTask(TASK_loadProject, meta)
    
    -- Get & load dependencies of the current project
    addTask(TASK_loadDependencies, meta)
    
    -- Setup file system & load WebRepo manifest
    local p = asset.documents.path .. "/webrepolib_cache"
    local fm = objc.NSFileManager.defaultManager
    if not fm:fileExistsAtPath_isDirectory_(p, true) then
        fm:createDirectoryAtPath_withIntermediateDirectories_attributes_error_(p, true, nil, nil)
    end
    
    -- Update the WebRepo manifest
    local function loadManifest(data)
        manifest = json.decode(data)
        for p, versions in pairs(manifest) do
            for _, v in ipairs(versions) do
                versions[v] = true
            end
        end
    end
    
    local manifestURL = "https://codeawebrepo.co.uk/manifest.json"
    local timedOut = false
    local function httpFail(err)
        -- Ignore if already loaded
        if manifest ~= nil then return end
        
        -- Read from disk instead?
        local data = readText(asset.documents.webrepolib_cache .. "manifest.json")
        if data == nil then
            error("Failed to GET " .. manifestURL .. " (" .. err .. ").")
        end
        print("Failed to update. Using cached manifest")
        loadManifest(data)
    end
    
    http.request(manifestURL, function(data, status)
        -- Ignore if already loaded
        if manifest ~= nil then return end
        
        if status ~= 200 then
            httpFail(tostring(status))
        else
            loadManifest(data)
            -- Save to disk
            saveText(asset.documents.webrepolib_cache .. "manifest.json", data)
        end
    end, httpFail)
    local reqTime = ElapsedTime
    
    -- Don't progress until the manifest is available
    print("Updating manifest")
    gate(function()
        -- Timeout?
        if (ElapsedTime - reqTime) > 2.0 and not timedOut then
            -- Read from disk if it's been downloaded already.
            -- If we don't already have it we'll continue waiting
            local data = readText(asset.documents.webrepolib_cache .. "manifest.json")
            if data then
                print("Timed out. Using cached manifest")
                loadManifest(data)
            end
            timedOut = true
        end
        
        -- Gate conditional on manifest being loaded
        return manifest ~= nil
    end)
    
    -- Success
    return true
end

TASK_validateProject = function(meta)
    if meta.name:match("^~") then
        meta.name = meta.name:sub(2) -- strip '~'
        meta.version = "__LOCAL__"
    else
        -- Get remote project
        
        -- Check project exists in manifest
        local versions = manifest[meta.name]
        if versions == nil then
            error("Project '" .. meta.name .. "' is not available.")
        end
        meta.version = meta.version or versions[#versions] -- Use latest by default
        if versions[meta.version] == nil then
            error("Project version '" .. meta.name .. "-" .. meta.version .. "' is not available.")
        end
        
        local metadata = readText(asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version .. "/metadata.json")
        if metadata == nil then
            local url = "https://codeawebrepo.co.uk/" .. socket.url.escape(meta.name) .. "/" .. socket.url.escape(meta.version) .. "/metadata.json"
            http.request(url, function(res)
                --print(res)
                saveText(asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version .. "/metadata.json", res)
                metadata = json.decode(res)
                meta.projName = metadata.name
                downloaded = true
            end, function(err)
                objc.warning("Failed to download metadata.json for '" .. meta.name .. ":" .. meta.version .. "' (" .. err .. ")")
                downloaded = true
            end)
            -- Wait until we receive the metadata file
            gate(function() return downloaded == true end)
        else
            metadata = json.decode(metadata)
            meta.projName = metadata.name
        end
        
        -- Check if the project is already installed
        if readText(asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version .. "/valid.txt") then
            -- Nothing to do!
            return true
        end
        
        -- Install the downloaded project
        addTask(TASK_installProject, meta)
    end
    
    -- Success
    return true
end

TASK_loadDependencies = function(meta)
    local wrl = {}
    
    if meta.version == "__LOCAL__" then
        -- Load the metadata file
        local fn = loadfile((asset.documents .. meta.name .. ".codea/WRL.lua").path, "t", wrl)
        if fn then
            if (WRL.CrashableCall(fn) == false) then
                return false
            end
        end
    else
        -- Remote project
        
        -- Load the metadata file
        local fn = loadfile((asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version .. "/" .. meta.name .. ".codea/WRL.lua").path, "t", wrl)
        if fn then
            if (WRL.CrashableCall(fn) == false) then
                return false
            end
        end
    end

    if wrl.wrlVersion and wrl.wrlVersion.required and wrl.wrlVersion.required > WRL.Version then
        error("Unable to load " .. meta.name .. " as the loaded WebRepoLib version is too old (v" .. tostring(WRL.Version) .. ' vs. v' .. tostring(wrl.wrlVersion.required) .. ')')
    end
    
    -- Allow overrides declared by the importer
    -- TODO: Use 'meta' tables to determine unique project loads.
    for k, v in pairs(wrl) do
        if meta[k] == nil then
            meta[k] = v
        end
    end
    
    -- Create a new env for the project to load into
    meta.env, meta.depenvs = createInternalEnv(meta)
    
    if meta.new_G == true then
        meta._G = meta.env
        
        -- Add some globals
        if meta.globals then
            for k,v in pairs(meta.globals) do
                meta._G[k] = v
            end
        end
    end
    meta.env._G = meta._G
    
    local function convertDepFormat(deps)
        if not deps then return end
        for i,dep in ipairs(deps) do
            
            -- Allow string only deps (not in a table)
            if type(dep) == "string" then
                dep = { dep }
                deps[i] = dep
            end
            
            local extra = dep[2] or {}
            dep.name = dep[1]
            for k,v in pairs(dep[2] or {}) do
                dep[k] = v
            end
            dep[1] = nil
            dep[2] = nil
        end
    end
    
    -- Convert to internal meta format
    convertDepFormat(meta.deps)
    convertDepFormat(meta.devdeps)
    
    local function loadDeps(deps)
        if not deps then return end
        for _,proj in ipairs(deps) do
            proj._G = meta._G
            addTask(TASK_loadProject, proj)
            addTask(TASK_loadDependencies, proj)
            addTask(TASK_validateProject, proj)
        end
    end
    
    -- Load any dependencies
    meta.deps = meta.deps or {}
    if meta.loaddev and meta.devdeps then
        table.move(meta.devdeps, 1, #meta.devdeps, #meta.deps+1, meta.deps)
    end
    loadDeps(meta.deps)
    
    -- Success
    return true
end

TASK_installProject = function(meta)
    local downloaded = false
    local ppath = (asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version).path
    
    -- Create project/version folder
    local fm = objc.NSFileManager.defaultManager
    if not fm:fileExistsAtPath_isDirectory_(ppath, true) then
        fm:createDirectoryAtPath_withIntermediateDirectories_attributes_error_(ppath, true, nil, nil)
    end
    
    local function install()
        print("Installing '" .. meta.name .. ":" .. meta.version .. "'")
        
        -- Extract .codea project from project.zip
        -- Open a new zip archive for reading
        local zip = Zip(asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version .. "/project.zip")
        
        -- Get all files in the zip
        local files = zip:listFiles()
        
        -- Extract all files
        for _,file in ipairs(files) do
            
            -- Ensure the dir exists
            local dir = ppath .. "/" .. file:match(".*/")
            if not fm:fileExistsAtPath_isDirectory_(dir, true) then
                fm:createDirectoryAtPath_withIntermediateDirectories_attributes_error_(dir, true, nil, nil)
            end
            
            -- Extract file
            local f = io.open(ppath .. "/" .. file, "wb")
            f:write(zip:readFile(file))
            f:close()
        end
        
        -- Close the zip file
        zip:close()
        
        -- Add 'valid' flag file
        saveText(asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version .. "/valid.txt", "READY_TO_USE")
    
        -- Wait until the cache folder is available
        gate(function()
            return asset.documents.webrepolib_cache[meta.name] ~= nil and asset.documents.webrepolib_cache[meta.name][meta.version] ~= nil
        end)
    end
    
    -- Get the project .zip
    local url = "https://codeawebrepo.co.uk/" .. socket.url.escape(meta.name) .. "/" .. socket.url.escape(meta.version) .. "/project.zip"
    
    -- Get the headers only first
    http.request(url, function(res, code, headers)
        
        -- Download files in chunks to avoid memory issues
        local size = math.tointeger(headers['Content-Length'])
        local totalSize = size
        local pos = 0
        local file = io.open(ppath .. "/project.zip", "wb")
        local CHUNK_SIZE = 1 * 1024 * 1024 -- 1MB at a time
        
        local function nextChunk()
            local chunkSize = math.min(size, CHUNK_SIZE)
            size = size - chunkSize
            http.request(url, function(res)
                file:write(res)
                pos = pos + chunkSize
                --if progress_cb then progress_cb(pos/totalSize) end
                if size > 0 then
                    nextChunk()
                else
                    file:close()
                    downloaded = true
                    addTask(install)
                end
            end, function(err)
                file:close()
                error("Failed to download '" .. meta.name .. ":" .. meta.version .. "' (" .. err .. ")")
            end, { headers = { Range = string.format("bytes=%d-%d", pos, pos + chunkSize - 1)}})
        end
        nextChunk()
        
    end, function(err)
        error("Failed to download '" .. meta.name .. ":" .. meta.version .. "' (" .. err .. ")")
    end, { method = "HEAD" })
    
    -- Wait until the project is installed
    print("Downloading '" .. meta.name .. ":" .. meta.version .. "'")
    gate(function() return downloaded == true end)
    
    -- Success
    return true
end

TASK_loadProject = function(meta)
    meta.version = meta.version or "__LOCAL__"
    -- Default value (not guaranteed to work if we fail to download the metadata file)
    meta.projName = meta.projName or meta.name
    
    -- TODO: Allow name overrides for use in 'require()'
    -- Use cached env if this project has already been loaded
    if ENVS[meta.name] and ENVS[meta.name][meta.version] ~= nil then
        local cmeta = ENVS[meta.name][meta.version]
        if cmeta.loadmain == meta.loadmain then
            meta.env = cmeta.env
            return true
        end
        -- TODO devdeps check too
    end
    
    local printName = meta.name .. ":" .. (((meta.version == "__LOCAL__") and "local") or meta.version)
    print("Loading '" .. printName .. "'")
    local ppath, plist
    if meta.version == "__LOCAL__" then
        -- Read plist
        plist = readText(asset.documents .. meta.name .. ".codea/Info.plist")
        if plist == nil then
            if (asset.documents .. meta.name .. ".codea") ~= nil then
                plist = "<key>Buffer Order</key><array><string>Main</string></array>"
            else
                error("Local project '" .. meta.name .. "' does not appear to be installed!")
            end
        end
        
        ppath = (asset.documents .. meta.name .. ".codea").path .. "/"
    else
        -- Remote project
        
        -- Read plist
        plist = readText(asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version .. "/" .. meta.projName .. ".codea/Info.plist")
        if plist == nil then
            if (asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version .. "/" .. meta.projName .. ".codea") ~= nil then
                plist = "<key>Buffer Order</key><array><string>Main</string></array>"
            else
                error("'" .. meta.name .. "' does not appear to be installed!", 0)
            end
        end
        ppath = (asset.documents.webrepolib_cache .. meta.name .. "/" .. meta.version .. "/" .. meta.projName .. ".codea").path .. "/"
    end
    
    -- Auto import dependencies
    -- (The dependency is copied into the current project's env)
    for _,dep in pairs(meta.deps) do
        if dep.export then
            WRL.Env.import(dep.env, meta.env)
        elseif dep.import or dep.import == nil then
            table.insert(meta.depenvs, dep.env)
        end
    end
    
    -- Load project into environment
    local buffers = plist:match("<key>Buffer Order</key>%s*<array>(.-)</array>")
    for b in buffers:gmatch("<string>(.-)</string>") do
        if (b == "Main" and meta.loadmain ~= true) or (b == "WRL") then
            goto CONTINUE
        end
        
        -- Load from file
        local fn, err = loadfile(ppath .. b .. ".lua", nil, meta.env)
        if fn == nil then
            error(err)
        end
                
        -- Execute file
        local success, err = WRL.CrashableCall(fn)
        if not success then
            -- Load failed so we should halt everything now
            WRL.TaskManager.batches = {}
            return false
        end
            
        ::CONTINUE::
    end
    
    -- Convert env variable to an external env (no metatable)
    meta.env = createExternalEnv(meta.env)
    
    -- Add env to ENVS table
    ENVS[meta.name] = ENVS[meta.name] or {}
    ENVS[meta.name][meta.version or "__LOCAL__"] = meta
    
    -- Success
    return true
end

TASK_go = function(meta)
    -- Nil internal refs
    WRL.Env = nil
    WRL.Tasks = nil
    
    -- Call OnLoad on every loaded project
    for _,versions in pairs(ENVS) do
        for _,meta in pairs(versions) do
            if meta.env.OnLoad then
                meta.env.OnLoad()
                -- Remove OnLoad() so it isn't repeated
                meta.env.OnLoad = nil
            end
        end
    end
    
    -- Import the root project to the global environment
    importEnv(meta.env)

    print("Done.") -- Celebrate?
    
    -- Run the setup function
    if type(setup) == "function" then setup() end
    
    -- Success
    return true
end

addTask(TASK_kickoff)

-- Override the require function to insert our
-- loaded dependencies.
local _require = require
function require(...)
    
    local args = {...}
    local name = args[1]
    local version = args[2]
    
    if name and ENVS[name] then
        if version and ENVS[name][version] then
            return ENVS[name][version][1]
        else
            for k,v in pairs(ENVS[name]) do
                return v[1]
            end
        end
    end
    
    return _require(...)
end

function import(...)
    local args = {...}
    local name = args[1]
    local version = args[2]
    
    local _cEnv -- env of caller
    local i = 2
    local info = debug.getinfo(i)
    while info ~= nil and _cEnv == nil do
        local func = info.func
        for i=1,info.nups do
            local k,v = debug.getupvalue(func, i)
            if k == "_ENV" then
                _cEnv = v
                break
            end
        end
        
        i = i + 1
        info = debug.getinfo(i)
    end
    
    -- Import the named project into the environment of the caller
    if name and ENVS[name] then
        if version and ENVS[name][version] then
            return WRL.Env.import(ENVS[name][version][1], _cEnv)
        else
            for k,v in pairs(ENVS[name]) do
                return WRL.Env.import(v[1], _cEnv)
            end
        end
    end
    
    error("Project not declared in WRL!")
end

-- Version 1
--------------------------------------------
function WRL.CacheProject(name, version, callback)
    print("Previewing " .. name .. "-" .. version)
    
    -- Non-sequential batch (tasks split across frames)
    local batch = WRL.TaskBatch(true)
    if callback then
        batch:add(callback)
    end

    -- Download and install the project
    local meta = {
        name = name,
        version = version
    }
    batch:add(TASK_validateProject, meta)

    -- Add the new batch
    WRL.TaskManager:addBatch(batch)
end

function WRL.LoadProject(name, version, globals, callback)
    ENVS = {}
    
    local meta = {
        name = name,
        version = version,
        loadmain = true,
        loaddev = false,
        new_G = true,
        globals = globals
    }
    
    local batch = WRL.TaskBatch(true)
    if callback then
        batch:add(function()
            -- Call OnLoad on every loaded project
            for _,versions in pairs(ENVS) do
                for _,meta in pairs(versions) do
                    if meta.env.OnLoad then
                        meta.env.OnLoad()
                        -- Remove OnLoad() so it isn't repeated
                        meta.env.OnLoad = nil
                    end
                end
            end
            -- Clear ENVS as we no longer need these
            ENVS = {}
            
            -- Return the environment via the callback
            callback(meta.env)
        end)
    end
    
    -- Load proj + dependencies
    batch:add(TASK_loadProject, meta)
    batch:add(TASK_loadDependencies, meta)
    batch:add(TASK_validateProject, meta)
    WRL.TaskManager:addBatch(batch)
end

function WRL.FeatureLevel(level)
    return level > WRL.Version
end

::EOF::